import datetime
import decimal
from types import FunctionType, MethodType, ModuleType
from google.protobuf.reflection import GeneratedProtocolMessageType
from google.protobuf.internal import api_implementation

if api_implementation.Type() == "upb":
    import google

import six

if six.PY2:
    from types import TypeType

try:
    import numpy
except ImportError:
    numpy = None

try:
    import torch
except ImportError:
    torch = None

try:
    import multidict
    multidict_inner = multidict._multidict
except ImportError:
    multidict = None
    multidict_inner = None


class NamespaceSerializerBase(object):
    BUILTIN_ATTRIBUTES_IGNORE = {
        '__dict__',
        '__module__',
        '__weakref__',
        '__name__',
        '__doc__',
        '__qualname__',
        '__spec__',
        '__defaults__',
        '__code__',
        '__globals__',
        '__closure__',
        '__annotations__',
        '__kwdefaults__',
        '__bases__'}

    try:
        BINARY_TYPES = (buffer, bytearray)
        CODE_TYPES = (FunctionType, MethodType, TypeType, ModuleType, six.MovedModule)
        PRIMITIVE_TYPES = (type(None), int, long, float, str, unicode, complex, decimal.Decimal) + BINARY_TYPES + CODE_TYPES + (datetime.datetime,)

    except NameError:
        BINARY_TYPES = (bytearray, bytes)
        CODE_TYPES = (FunctionType, MethodType, type, ModuleType, six.MovedModule)
        PRIMITIVE_TYPES = (type(None), int, float, str, complex, decimal.Decimal) + BINARY_TYPES + CODE_TYPES + (datetime.datetime,)

    def __init__(self, use_string_cache=False):
        self.use_string_cache = use_string_cache
        self.string_cache = {}

        if use_string_cache:
            # Lock the 0 index since some variant will have no originalType (container for example)
            self.string_cache[""] = 0

        self.estimated_pending_bytes = 0

    def get_string_cache(self):
        return self.string_cache

    def get_estimated_pending_bytes(self):
        return self.estimated_pending_bytes

    def _get_string_index_in_cache(self, string_to_cache):
        index = self.string_cache.get(string_to_cache, None)
        if index is None:
            index = len(self.string_cache)
            # We estimate each character is one byte in utf-8 and overhead is 5 bytes
            self.estimated_pending_bytes += len(string_to_cache) + 5
            self.string_cache[string_to_cache] = index

        return index

    @staticmethod
    def _get_object_width(obj):
        object_width = 0
        if hasattr(obj, '__dict__') and obj.__dict__:
            object_width += len(obj.__dict__)
        if hasattr(obj, '__slots__') and obj.__slots__:
            object_width += len(obj.__slots__)
        return object_width

    if six.PY2:
        @staticmethod
        def normalize_string(obj):
            if isinstance(obj, str):
                return unicode(obj, errors="replace")
            else:
                return unicode(obj)
    else:
        @staticmethod
        def normalize_string(obj):
            return obj

    @staticmethod
    def is_numpy_obj(obj):
        if not numpy:
            return False

        return not isinstance(obj, six.MovedModule) and isinstance(obj, numpy.generic)

    @staticmethod
    def is_torch_obj(obj, obj_type=None):
        if torch is None:
            return False

        if obj_type is None:
            obj_type = type(obj)

        module = getattr(obj_type, '__module__', None)
        if not module:
            return False

        return module.startswith('torch')

    @staticmethod
    def is_multidict_obj(obj):
        if not multidict_inner:
            return False

        return isinstance(obj, (multidict_inner.MultiDict, multidict_inner.CIMultiDict, multidict_inner.MultiDictProxy, multidict_inner.CIMultiDictProxy))

    @staticmethod
    def is_protobuf_obj(obj):
        klass = getattr(obj, '__class__', None)
        if api_implementation.Type() == "upb":
            if klass is not None and isinstance(klass, google._upb._message.MessageMeta):
                return True
        else:
            if klass is not None and isinstance(klass, GeneratedProtocolMessageType):
                return True

        return False



import six

from .dumped_object_namespace import DumpedObjectNamespace

from .python_object_namespace import PythonObjectNamespace

from rook.exceptions import RookKeyNotFound

LIST_TYPE = (list, tuple, set, frozenset)
try:
    import numpy
    LIST_TYPE = LIST_TYPE + (numpy.ndarray, )
except ImportError:
    numpy = None


class CollectionNamespace(DumpedObjectNamespace):

    def __init__(self, original_size, type, common_type, attributes):
        super(CollectionNamespace, self).__init__(type, common_type, attributes, self.METHODS)
        self.original_size = original_size

    def read_key(self, key):
        try:
            return self[key]
        except (KeyError, IndexError):
            raise RookKeyNotFound(key)

    def size(self, args):
        return PythonObjectNamespace(len(self))

    def original_size(self, args):
        return PythonObjectNamespace(self.original_size)

    METHODS = (size, original_size)


class DictNamespace(CollectionNamespace, dict):

    def __init__(self, collection, original_size=None, collection_type=None, attributes={}):
        if not collection_type:
            collection_type = str(type(collection))

        if not original_size:
            original_size = len(collection)

        dict.__init__(self, collection)
        CollectionNamespace.__init__(self, original_size, collection_type, u'dict', attributes)

    def to_dict(self):
        items = [(key.to_dict(), value.to_dict()) for key, value in six.iteritems(self)]

        return {
            u'@namespace': self.__class__.__name__,
            u'@common_type': self.common_type,
            u'@original_type': self.type,
            u'@original_size': self.original_size,
            u'@attributes': self.get_attributes_dict(),
            u'@value': items
            }

    def to_simple_dict(self):
        items = {}
        for key, value in six.iteritems(self):
            items[key.obj] = value.to_simple_dict()

        return items


class ListNamespace(CollectionNamespace, list):

    def __init__(self, collection, original_size=None, collection_type=None, common_type=None, attributes={}):
        if not original_size:
            original_size = len(collection)

        if not collection_type:
            collection_type = str(type(collection))

        if not common_type:
            common_type = self.get_common_type(collection)

        list.__init__(self, collection)
        CollectionNamespace.__init__(self, original_size, collection_type, common_type, attributes)

    @staticmethod
    def is_numpy_obj(obj):
        if not numpy:
            return False

        return isinstance(obj, numpy.ndarray)

    @staticmethod
    def get_common_type(collection):
        if isinstance(collection, list):
            return u'list'
        if ListNamespace.is_numpy_obj(collection):
            return u'array'
        elif isinstance(collection, tuple):
            return u'tuple'
        elif isinstance(collection, (set, frozenset)):
            return u'set'
        else:
            return u'list_unknown'

    def __hash__(self):
        return hash(tuple(k for k in self))

    def __eq__(self, other):
        if type(self) != type(other):
            return False

        return tuple(k for k in self).__eq__(tuple(k for k in other))

